// datamodifier.C

/******************************************************************************
 *
 *  MiXViews - an X window system based sound & data editor/processor
 *
 *  Copyright (c) 1993, 1994 Regents of the University of California
 *
 *  Author:     Douglas Scott
 *  Date:       December 13, 1994
 *
 *  Permission to use, copy and modify this software and its documentation
 *  for research and/or educational purposes and without fee is hereby granted,
 *  provided that the above copyright notice appear in all copies and that
 *  both that copyright notice and this permission notice appear in
 *  supporting documentation. The author reserves the right to distribute this
 *  software and its documentation.  The University of California and the author
 *  make no representations about the suitability of this software for any 
 *  purpose, and in no event shall University of California be liable for any
 *  damage, loss of data, or profits resulting from its use.
 *  It is provided "as is" without express or implied warranty.
 *
 ******************************************************************************/


#ifdef __GNUG__
#pragma implementation
#endif

#include "application.h"
#include "localdefs.h"
#include "controller.h"
#include "datamodifier.h"
#include "data.h"
#include "editor.h"
#include "formatrequester.h"
#include "lpcdata.h"
#include "processfun.h"
#include "range.h"
#include "request.h"
#include "sound.h"
#include "valuerequester.h"

DataModifier::DataModifier(Data* data) : myData(data), myEditor(nil) {
	myData->ref();
}

DataModifier::DataModifier(Data* data, DataEditor* editor)
		: myData(data), myEditor(editor) {
	myData->ref();
}

DataModifier::~DataModifier() {
	Resource::unref(myData);
}

int
DataModifier::apply() {
	int status = false;
	if(ok()) {
		Application::inform(message());
		status = doApply(myData);
	}
	return status;
}

void
DataModifier::doPostModifyCommands() {
	if(myEditor != nil)
		doEditorCommands(myEditor);
}

int
DataModifier::sampRate() { return myData->sRate(); }

//********

Modifier *
Reverser::create(DataEditor* ed) {
	return new Reverser(ed->currentSelection());
}

int
Reverser::doApply(Data* data) {
	data->reverse();
	return true;
}

Modifier *
Reverser::createUnDo() {
	return new Reverser(target());
}

//********

Modifier *
ByteSwapper::create(DataEditor* ed) {
	return new ByteSwapper(ed->currentSelection());
}

int
ByteSwapper::doApply(Data* data) {
	data->swapBytes();
	return true;
}

Modifier *
ByteSwapper::createUnDo() {
	return new ByteSwapper(target());
}

//********

class NormalizeRequester : public TitledRequester {
	friend class Normalizer;
protected:
	NormalizeRequester(Normalizer *n)
		: TitledRequester("Normalize Current Selection:"), client(n) {}
	redefined void configureRequest(Request *);
	redefined boolean confirmValues();
private:
	Normalizer* client;
};

void
NormalizeRequester::configureRequest(Request *request) {
	request->appendLabel("This will rescale the current selection's amplitudes");
	request->appendLabel("to fit between -1 and 1.");
}

boolean
NormalizeRequester::confirmValues() {
	return (
		(client->target()->dataType() > IntData)
		||
		Application::confirm(
			"Warning:  Performing this operation on non-floating-point samples",
			"will irreversibly destroy the data.",
			"Continue anyway?",
			Cancel
		)
	);
}

//********

Modifier *
Normalizer::create(DataEditor* ed) {
	return new Normalizer(ed->currentSelection());
}

Requester *
Normalizer::createRequester() {
	return new NormalizeRequester(this);
}

int
Normalizer::doApply(Data* data) {
	reverseFactor = data->maxValue();		// store inverse for undo
	data->normalize();
	return true;
}

// undo is only possible when data type is float or double

Modifier *
Normalizer::createUnDo() {
	return (target()->dataType() >= FloatData) ?
		new Scaler(target(), reverseFactor) : nil;
}

//********

class TimeInsertRequester : public TitledRequester {
	friend class SpaceInsert;
protected:
	TimeInsertRequester(SpaceInsert *);
	redefined void configureRequest(Request *);
	redefined boolean confirmValues();
private:
	SpaceInsert* client;
	double timeToInsert;
};

TimeInsertRequester::TimeInsertRequester(SpaceInsert* s)
	: TitledRequester("Insert Blank Space:"),
	  client(s), timeToInsert(client->_nFrames / client->sampRate()) {}

void
TimeInsertRequester::configureRequest(Request* request) {
	request->appendValue("Amount To Insert (seconds):",
	                     &timeToInsert, NonNegativeNumbers);
}

boolean
TimeInsertRequester::confirmValues() {
	client->_nFrames = round(timeToInsert * client->sampRate());
	return true;
}

//********

int SpaceInsert::_savedFrames = 1;

SpaceInsert::SpaceInsert(Data* d, int nframes)
		: DataModifier(d), _nFrames(nframes), _insertTime(false) {
	initialize();
}

SpaceInsert::SpaceInsert(Data* d, boolean asTime)
	: DataModifier(d), _nFrames(d->length()), _insertTime(asTime) {}

Modifier *
SpaceInsert::create(DataEditor* ed) {
	return new SpaceInsert(ed->currentSelection());
}

Requester *
SpaceInsert::createRequester() {
	if(_insertTime)
		return new TimeInsertRequester(this);
	else
		return new RangedValueRequester<int>(
			"Insert Blank Space:", "Amount To Insert (frames):",
			this->_nFrames, NonNegativeIntegers);
}

void
SpaceInsert::saveConfig() {
    _savedFrames = _nFrames;
}

int
SpaceInsert::doApply(Data *data) {
	Range edit(0, _nFrames);
	boolean was = Data::deferRescan(true);	// inserting zeros so no rescan
	int status = data->spliceIn(edit);
	Data::deferRescan(was);
	return status;
}

Modifier *
SpaceInsert::createUnDo() {
	return new OutSplicer(target(), Range(0, _nFrames));
}

//********

OutSplicer::OutSplicer(Data* d, const Range& r) : DataModifier(d), region(r) {
	initialize();
}

Modifier *
OutSplicer::create(DataEditor* de) {
	return nil;
}

int
OutSplicer::doApply(Data* data) {
	if(region == data->frameRange()) {
		Application::alert("Entire file may not be spliced out.",
			"Select only a portion of the file.");
		return false;
	}
	else
		return data->spliceOut(region);
}

Modifier *
OutSplicer::createUnDo() {
	return nil;
}

//********

class LengthChangeRequester : public TitledRequester {
	friend class LengthChanger;
protected:
	LengthChangeRequester(const char* title, LengthChanger* l)
		: TitledRequester(title), client(l) {}
	redefined void configureRequest(Request *);
	redefined boolean confirmValues();
protected:
	LengthChanger* client;
};

void
LengthChangeRequester::configureRequest(Request* request) {
	Range lengths = PositiveIntegers;
	lengths.set(8, lengths.intMax());	// min length 8
	request->appendValue("New Length (in frames):",
	                     &client->newLength, lengths);
	request->appendLabel("(Minimum duration is 8 frames)");
}

boolean
LengthChangeRequester::confirmValues() {
	boolean status = false;
	status = (client->newLength >= client->oldLength ||
		Application::confirm(
			"This operation will destroy any data past",
			"the new file endpoint, and is not reversible.",
			"Continue?",
			Cancel)
	);
	return status;
}

//********

class DurChangeRequester : public LengthChangeRequester {
	friend class LengthChanger;
protected:
	DurChangeRequester(const char* title, LengthChanger *);
	redefined void configureRequest(Request *);
	redefined boolean confirmValues();
private:
	double newDuration;
};

DurChangeRequester::DurChangeRequester(const char* title, LengthChanger* client)
		: LengthChangeRequester(title, client) {
	newDuration = double(client->newLength)/client->sampRate();
}

void
DurChangeRequester::configureRequest(Request* request) {
	Range durations = AllNumbers;
	durations.set(8.0/client->sampRate(), durations.max());
	request->appendValue("New Duration (seconds):", &newDuration,
			     durations);
	request->appendLabel("(Minimum duration is the time required for 8 frames)");
}

boolean
DurChangeRequester::confirmValues() {
	client->newLength = round(newDuration * client->sampRate());
	return LengthChangeRequester::confirmValues();
}

//********

int LengthChanger::_savedNewLength = 0;

LengthChanger::LengthChanger(Data* d, boolean asDur)
	: DataModifier(d),
	  oldLength(d->length()),
	  newLength(_savedNewLength == 0 ? oldLength : _savedNewLength),
	  asDuration(asDur) {}

LengthChanger::LengthChanger(Data* d, int newlen)
		: DataModifier(d),
		  oldLength(d->length()), newLength(newlen), asDuration(false) {
	initialize();
}

Modifier *
LengthChanger::create(DataEditor* ed) {
	return new LengthChanger(ed->model());
}

Requester *
LengthChanger::createRequester() {
	if(asDuration)
		return new DurChangeRequester("Change File Duration:", this);
	else
		return new LengthChangeRequester("Change File Length:", this);
}

void
LengthChanger::saveConfig() {
    _savedNewLength = newLength;
}

int
LengthChanger::doApply(Data *data) {
	int status = data->changeLength(newLength);
	if(status == true) {
		// defer rescan if increasing length since extending with zeros
		boolean was = Data::deferRescan(newLength >= oldLength);
		data->Notify();		// data does not automatically do this 
		Data::deferRescan(was);
	}
	return status;
}

Modifier *
LengthChanger::createUnDo() {
	return (newLength > oldLength) ?
		new LengthChanger(target(), oldLength) : nil;
}

//********

class SampRateRequester : public TitledRequester {
	friend class SampRateChanger;
protected:
	SampRateRequester(SampRateChanger *s)
		: TitledRequester("Change Sampling Rate:"), client(s) {}
	redefined void configureRequest(Request *);
	redefined boolean confirmValues();
private:
	SampRateChanger* client;
};

void
SampRateRequester::configureRequest(Request* request) {
	request->appendValue("New Sampling Rate:",
	                     &client->newSrate, PositiveIntegers);
	if(client->target()->isA(Sound_Data))
		request->appendChoice(
			"Interpolate sound to preserve pitch level:",
			"|No|Yes|", &client->doInterp, true);
	else client->doInterp = False;
}

boolean
SampRateRequester::confirmValues() {
	if(client->doInterp == Requester::True)
		return Application::confirm("This operation is not reversible.",
			"Continue?");
	else
		return true;
}

//********

int SampRateChanger::_savedNewSrate = 0;
ChoiceValue SampRateChanger::_savedDoInterp = Requester::True;

SampRateChanger::SampRateChanger(Data* d)
	: DataModifier(d),
	  newSrate(_savedNewSrate == 0 ? sampRate() : _savedNewSrate),
	  doInterp(_savedDoInterp) {}
		
SampRateChanger::SampRateChanger(Data* d, int newsrate, boolean interp)
		: DataModifier(d), newSrate(newsrate),
		doInterp(interp ? Requester::True : Requester::False) {
	initialize();
}

Modifier *
SampRateChanger::create(DataEditor* ed) {
	return new SampRateChanger(ed->model());
}

Requester *
SampRateChanger::createRequester() {
	return new SampRateRequester(this);
}

void
SampRateChanger::saveConfig() {
    _savedNewSrate = newSrate;
    _savedDoInterp = doInterp;
}

int
SampRateChanger::doApply(Data *data) {
	return data->changeSRate(newSrate, doInterp == Requester::True);
}

//********

class FormatChangeRequester : public FormatRequester {
	friend class FormatChanger;
protected:
	FormatChangeRequester(FormatChanger *);
	redefined boolean confirmValues();
private:
	FormatChanger* client;
};

FormatChangeRequester::FormatChangeRequester(FormatChanger* f)
	: FormatRequester(nil, "Convert format to:", f->newType),
	  TitledRequester("Change Sound Sample Format:"),
	  client(f) {}

boolean
FormatChangeRequester::confirmValues() {
	boolean status = true;
	if(client->newType == client->oldType) {
		Application::alert("Samples are already in this format.");
		status = false;
	}
	return status;
}

//********

FormatChanger::FormatChanger(Data* d)
		: DataModifier(d), oldType(d->dataType()) {
	newType = (oldType == ShortData) ? FloatData : ShortData;
}
		
FormatChanger::FormatChanger(Data* d, DataType newtype)
		: DataModifier(d), oldType(d->dataType()), newType(newtype) {
	initialize();
}

Modifier *
FormatChanger::create(DataEditor* ed) {
	return new FormatChanger(ed->model());
}

Requester *
FormatChanger::createRequester() {
	return new FormatChangeRequester(this);
}

int
FormatChanger::doApply(Data *data) {
	return ((Sound *) data)->changeDataType(DataType(newType));
}

Modifier *
FormatChanger::createUnDo() {
	return (newType > oldType) ?
		new FormatChanger(target(), DataType(oldType)) : nil;
}

//********

class AutoCorrelateRequester : public TitledRequester {
	friend class AutoCorrelator;
protected:
	AutoCorrelateRequester(AutoCorrelator* a)
		: TitledRequester("Auto-correlate Data (remove glitches):"), _client(a) {}
	redefined void configureRequest(Request *);
	redefined boolean confirmValues();
protected:
	AutoCorrelator* _client;
};

void
AutoCorrelateRequester::configureRequest(Request* request) {
	request->appendValue("Deviation threshold:",
	                     &_client->_slopeThreshold, NonNegativeNumbers);
	request->appendValue("Number of frames to look ahead:",
	                     &_client->_lookAhead, Range(4, 64), True);
}

boolean
AutoCorrelateRequester::confirmValues() {
	boolean status = true;
	if (_client->_slopeThreshold == 0.0) {
		Application::alert("Threshold must be factor greater than zero.");
		status = false;
	}
	if (_client->_lookAhead >= _client->target()->length() / 2) {
		Application::alert("Selected area too short for this look-ahead.",
						   "Extend your selected area to the right");
		status = false;
	}
	return status;
}

//********

float AutoCorrelator::_savedSlopeThreshold = 5.0;
int AutoCorrelator::_savedLookAhead = 4;

AutoCorrelator::AutoCorrelator(Data* d, float slopeThreshold, int lookAhead)
	: DataModifier(d), _slopeThreshold(slopeThreshold), _lookAhead(lookAhead)
{
	initialize();
}

AutoCorrelator::AutoCorrelator(Data* d)
	: DataModifier(d),
	_slopeThreshold(_savedSlopeThreshold), _lookAhead(_savedLookAhead)
{
}

Modifier *
AutoCorrelator::create(DataEditor *ed)
{
	return new AutoCorrelator(ed->model());
}

Modifier *
AutoCorrelator::createUnDo()
{
	return NULL;
}

Requester *
AutoCorrelator::createRequester()
{
	return new AutoCorrelateRequester(this);
}

void
AutoCorrelator::saveConfig() {
    _savedSlopeThreshold = _slopeThreshold;
    _savedLookAhead = _lookAhead;
}

int
AutoCorrelator::doApply(Data *data)
{
	const int frameSize = _lookAhead + 1;
	double *frame = new double[frameSize * 2];
	Data *dataFrame;
	double sum = 0.0;
	int count = 0;
	int endFrame = data->length() - 1;
	Range segment(0, 2 * frameSize - 1);
	while (segment.intMax() <= endFrame)
	{
//		printf("cloning segment %d, %d\n", segment.intMin(), segment.intMax());
		dataFrame = data->clone(segment);
		int got = dataFrame->getArray(frame, frameSize * 2);
//		printf("getArray returned %d out of %d frames\n", got, frameSize * 2);
		int toScan = min(frameSize, got);
		int lookEnd = got;
//		printf("scanning segment %d, %d\n", segment.intMin(), segment.intMin() + toScan - 1);
		for (int n = 0; n < toScan - 1; ++n, count)
		{
			++count;
			double slopeAverage = sum / (count - 1);
			double cur = frame[n];
			double next = frame[n+1];
			double absSlope = abs(next - cur) / cur;
//			printf("\tpair %d: [%g, %g] slopeAverage: %g  current slope: %g\n",
//				   n, cur, next, slopeAverage, absSlope);
			if (slopeAverage > 0.0 && absSlope / slopeAverage > _slopeThreshold)
			{
				int lastGoodIndex = n;
				int idx;
//				printf("\tlooking ahead from slot %d.  Last good value: %g\n",
//					  lastGoodIndex + 1, cur);
				// We hit a glitch -- search ahead for a good sample.
				for (idx = n+2; idx < lookEnd; ++idx)
				{
					// The threshold against which we test gets wider the farther 
					// we have to look ahead.
					double lookThreshold = _slopeThreshold * double(idx - n) / 2.0;
					double lookValue = frame[idx];
					double testSlope = abs(lookValue - cur) / cur;
//					printf("\t\tpair %d: [%g, %g] testSlope: %g  threshold: %g\n",
//					       idx, cur, lookValue, testSlope, lookThreshold);
					if (testSlope / slopeAverage < lookThreshold)
					{
						// OK, found a good one.  Now adjust all samples
						// between lastGoodIndex and where we are now.
//						printf("\t\tfilling between slot %d [%g] and slot %d [%g]\n",
//							   n, cur, idx, lookValue);
						for (int x = n+1; x < idx; ++x, ++count)
						{
							double fraction = double(x - n) / (idx - n);
							double newVal = cur + fraction * (lookValue - cur);
//							printf("\t\t\tresetting slot %d with %g\n",
//								   x, newVal);
							dataFrame->set(newVal, x);
						}
						sum += testSlope;
//						printf("\t\tbreaking to outer loop.\n");
						break;
					}
				}
				if (idx == lookEnd)
				{
					Application::alert("Lookahead failed to find target within slope range.",
									   "Run again with larger lookahead or slope factor");
					delete [] frame;
					data->Notify();
					return Succeed;
				}
				n = idx - 1;	// Overlap the return to the pair search by one
			}
			else
			{
				sum += absSlope;
			}
		}
		Resource::unref(dataFrame);
		segment += (frameSize - 1);		// each scan overlaps last by one frame
	}
	delete [] frame;
	data->Notify();
	return Succeed;
}

//********

Modifier *
FrameStabilizer::create(DataEditor* ed) {
	return new FrameStabilizer(ed->model());
}

int
FrameStabilizer::applyToLPCData(LPCData* lpc) {
	lpc->stabilizeFrames();
	return true;
}
